from flask import current_app, abort
from frasco.ext import get_extension_state
from itsdangerous import URLSafeTimedSerializer, BadSignature
from werkzeug.local import LocalProxy


__all__ = ('user_token_serializer', 'generate_user_token', 'read_user_token', 'read_user_token_or_404')


TOKEN_NS_ACCESS_TOKEN = 'access-token'
TOKEN_NS_2FA = '2fa'
TOKEN_NS_PASSWORD_RESET = 'password-reset'
TOKEN_NS_VALIDATE_EMAIL = 'validate-email'


def get_user_token_serializer():
    return URLSafeTimedSerializer(current_app.config['SECRET_KEY'])


user_token_serializer = LocalProxy(get_user_token_serializer)


def get_token_identifier_property():
    return getattr(get_extension_state('frasco_users').Model, '__token_identifier_property__', 'id')


def generate_user_token(user, salt=None):
    return user_token_serializer.dumps(getattr(user, get_token_identifier_property()), salt=salt)


def read_user_token(token, salt=None, max_age=None):
    try:
        user_id = user_token_serializer.loads(token, salt=salt, max_age=max_age)
        return model.query.filter(getattr(get_extension_state('frasco_users').Model, get_token_identifier_property()) == user_id).first()
    except BadSignature:
        return None


def read_user_token_or_404(*args, **kwargs):
    user = read_user_token(*args, **kwargs)
    if not user:
        abort(404)
    return user



from flask import after_this_request, _request_ctx_stack
from flask_sqlalchemy import SignallingSession
from frasco.ctx import ContextStack, DelayedCallsContext
from frasco.utils import AttrDict
from contextlib import contextmanager
from sqlalchemy import event
import functools
from .ext import db
import logging


__all__ = ('transaction', 'as_transaction', 'current_transaction', 'is_transaction', 'delayed_tx_calls', 'after_transaction_commit')


_transaction_ctx = ContextStack(default_item=True)
delayed_tx_calls = DelayedCallsContext()
logger = logging.getLogger('frasco.models')


@contextmanager
def transaction():
    if not _transaction_ctx.top:
        logger.debug('BEGIN TRANSACTION')
    _transaction_ctx.push()
    delayed_tx_calls.push()
    try:
        yield
        _transaction_ctx.pop()
        if not _transaction_ctx.top:
            logger.debug('COMMIT TRANSACTION')
            db.session.commit()
        else:
            db.session.flush()
        delayed_tx_calls.pop()
    except:
        _transaction_ctx.pop()
        if not _transaction_ctx.top:
            logger.debug('ROLLBACK TRANSACTION')
            db.session.rollback()
        delayed_tx_calls.pop(drop_calls=True)
        raise


def is_transaction():
    return bool(_transaction_ctx.top)


def as_transaction(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with transaction():
            return func(*args, **kwargs)
    return wrapper


_current_transaction_ctx = ContextStack()
current_transaction = _current_transaction_ctx.make_proxy()


def after_transaction_commit(func):
    delayed_tx_calls.call(func, [], {})


@event.listens_for(SignallingSession, 'after_begin')
def on_after_begin(session, transaction, connection):
    _current_transaction_ctx.push(AttrDict())

@event.listens_for(SignallingSession, 'after_commit')
def on_after_commit(session):
    _current_transaction_ctx.pop()

@event.listens_for(SignallingSession, 'after_rollback')
def on_after_rollback(session):
    _current_transaction_ctx.pop()
